%% Szerző:  Páter-Részeg Attila
%% NEPTUN:  RD1SET
%% Eseménykezelés Erlangban

%% Használt típusok:
%% szerzo() = string() | atom()
%% konyvcim() = string() | atom()
%% darabszam() = integer()

-module(adatbazis).
-import(lists, [foreach/2]).

-export([init/1, terminate/2]).
-export([handle_event/2, handle_call/2, handle_info/2, code_change/3]).

-include_lib("stdlib/include/qlc.hrl").

%% A könyv tábla egy rekordja
-record(konyv, {szerzo, konyvcim, akt_db, osszes_db}).

%% ============================================================================================================================
%% Eseménykezelők

%% init(State) -> {ok, State}
%% Az eseménykezelő hozzáadásakor automatikusan meghívódik, létrehozza az adatbázist, ha az még nem létezik
init([]) ->
  mnesia:create_schema([node()]),
  mnesia:start(),
  mnesia:create_table(konyv,[ {disc_copies, [node()] },{attributes, record_info(fields,konyv)} ]),
  {ok, []}.

%% handle_event({atom(), szerzo(), konyvcim(), darabszam()}, State) -> {ok, State}
%% kezeljük a kölcsönzés és visszahozás eseményét 
handle_event({visszahoz, Szerzo, KonyvCim, Db}, State) ->

  % lekérdezzük a kért könyvhöz tartozó darabszámokat
  {AktualisDarabszam, Osszes} = bent_levo_db(Szerzo, KonyvCim),

  % Ha lehetséges a visszahozás, töröljük arégi bejegyzést, és beszúrjuk a friss verziót
  if
    Osszes > 0, AktualisDarabszam < Osszes, Osszes >= AktualisDarabszam + Db ->
      konyv_torlese(Szerzo, KonyvCim),
      uj_konyv(Szerzo, KonyvCim, Db + AktualisDarabszam, Osszes),
      io:format("[SIKER] A ~p:~p konyv(ek) leadva.~n", [Szerzo, KonyvCim]);
    true -> 
      io:format("[SIKERTELEN] A ~p:~p cimu konyv nem szerepel a nyilvantartasban, vagy nem hozhato vissza.~n", [Szerzo, KonyvCim])
    end,    
  {ok, State};
handle_event({kolcsonoz, Szerzo, KonyvCim, Db}, State) ->
    {AktualisDarabszam, Osszes} = bent_levo_db(Szerzo, KonyvCim),
    % Ha lehetséges a kölcsönzés, töröljük a régi bejegyzést, és beszúrjuk a módosítottat
    if
      Osszes > 0, AktualisDarabszam >= Db -> 
        konyv_torlese(Szerzo, KonyvCim),
        uj_konyv(Szerzo, KonyvCim, AktualisDarabszam - Db, Osszes),
        io:format("[SIKER] A ~p:~p konyv(ek) kiveve.~n", [Szerzo, KonyvCim]);
      true -> io:format("[SIKERTELEN] Nincs ~p:~p cimu konyv a konyvtarban.~n", [Szerzo, KonyvCim])
    end,
    {ok, State};
handle_event(_, Pid) ->
    {ok, Pid}.

%% handle_call(atom(), State) -> {ok, Reply, State}
%% A specifikus kérések kezelése
handle_call(beolvasas, State) ->
    {ok, "Adatbazis betoltve!", State};
handle_call({uj_konyv, {S, K, D}}, State) ->
    uj_konyv(S, K, D, D),
    {ok, "Uj konyv sikeresen tarolva!", State};
handle_call(in, State) ->
    Eredmeny = bent_levo_konyvek(),
    {ok, Eredmeny, State};
handle_call(out, State) ->
    Eredmeny = kint_levo_konyvek(),
    {ok, Eredmeny, State};
handle_call(all, State) ->
    Eredmeny = osszes_konyv(),
    {ok, Eredmeny, State};
handle_call(_, State) ->
    {ok, ok, State}.

%% Jelenleg nem használt függvények
handle_info(_, State) ->
    {ok, State}.

code_change(_, State, _) ->
    {ok, State}.

%% Az eseménykezelő leválasztásakor meghívódó függvény
terminate(_, _) ->
    ok. 

%% ============================================================================================================================
%% Privát függvények

%% A bent lévő könyvek darabszámának kilistázása
bent_levo_db(S, K) ->
  Darab = mnesia:transaction( 
  fun() ->
    qlc:eval( qlc:q([{X#konyv.akt_db, X#konyv.osszes_db} || X <- mnesia:table(konyv), X#konyv.szerzo == S, X#konyv.konyvcim == K])) 
  end),
  case Darab of
    {atomic,[{A,B}]} -> {A,B};
    _ -> {0,0}
  end.

%% A bent lévő, kölcsönözhető könyvek kilistázása
bent_levo_konyvek() ->
  mnesia:transaction( 
  fun() ->
    qlc:eval( qlc:q([{X#konyv.szerzo, X#konyv.konyvcim, X#konyv.akt_db} || X <- mnesia:table(konyv), X#konyv.akt_db =/= 0])) 
  end).

%% A kikölcsönzött könyvek kilistázása
kint_levo_konyvek() ->
  mnesia:transaction( 
  fun() ->
    qlc:eval( qlc:q([{X#konyv.szerzo, X#konyv.konyvcim, X#konyv.osszes_db - X#konyv.akt_db} || X <- mnesia:table(konyv), X#konyv.akt_db =/= X#konyv.osszes_db])) 
  end).

%% A nyilvántartásban szereplő összes könyv kilistázása
osszes_konyv() ->
  mnesia:transaction( 
  fun() ->
    qlc:eval( qlc:q([{X#konyv.szerzo, X#konyv.konyvcim, X#konyv.akt_db, X#konyv.osszes_db} || X <- mnesia:table(konyv)])) 
  end).

%% Megadott könyvhöz tartozó sor törlése az adatbázisból
konyv_torlese(S, K) ->
  Id = {konyv, S, K},
  F = fun() ->
    mnesia:delete(Id)
  end,
  mnesia:transaction(F).

%% Új könyv felvétele az adatbázisba
uj_konyv(S, K, D, O) ->
  Fun = fun() ->
    mnesia:write(
    #konyv{szerzo=S,konyvcim=K, akt_db=D, osszes_db=O} )
  end,
  mnesia:transaction(Fun).